@@ -1,6 +1,6 @@
 # pywws - Python software for USB Wireless Weather Stations
 # http://github.com/jim-easterbrook/pywws
-# Copyright (C) 2008-18  pywws contributors
+# Copyright (C) 2008-19  pywws contributors
 
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -69,6 +69,7 @@ Detailed API
 
 import csv
 from datetime import date, datetime, timedelta, MAXYEAR
+import logging
 import os
 import sys
 import time
@@ -76,12 +77,12 @@ import time
 from pywws.constants import DAY
 from pywws.weatherstation import WSDateTime, WSFloat, WSInt, WSStatus
 
+logger = logging.getLogger(__name__)
 
 
 class _Cache(object):
     def __init__(self):
         self.data = []
-        self.ptr = 0
         self.path = ''
         self.lo = date.max
         self.hi = date.min
@@ -89,33 +90,34 @@ class _Cache(object):
 
     def copy(self, other):
         self.data = other.data
-        self.ptr = other.ptr
         self.path = other.path
         self.lo = other.lo
         self.hi = other.hi
         self.dirty = False
 
-    def set_ptr(self, idx):
+    def get_ptr(self, idx):
+        """Return index at which to insert a record with timestamp idx
+        and boolean indicating if there is already data with that
+        timestamp.
+
+        """
         hi = len(self.data) - 1
-        if hi < 0 or self.data[0]['idx'] >= idx:
-            self.ptr = 0
-            return
+        if hi < 0 or self.data[0]['idx'] > idx:
+            return 0, False
+        if self.data[0]['idx'] == idx:
+            return 0, True
         if self.data[hi]['idx'] < idx:
-            self.ptr = hi + 1
-            return
+            return hi + 1, False
         lo = 0
-        start = min(self.ptr, hi)
-        if self.data[start]['idx'] < idx:
-            lo = start
-        else:
-            hi = start
-        while hi > lo + 1:
+        while hi - lo > 1:
             mid = (lo + hi) // 2
             if self.data[mid]['idx'] < idx:
                 lo = mid
+            elif self.data[mid]['idx'] == idx:
+                return mid, True
             else:
                 hi = mid
-        self.ptr = hi
+        return hi, self.data[hi]['idx'] == idx
 
 
 class CoreStore(object):
@@ -183,23 +185,21 @@ class CoreStore(object):
         a, b = self._slice(i)
         if a > b:
             return
+        # use separate cache as something might change self._rd_cache
+        # during yield
+        cache = _Cache()
         # go to start of slice
-        self._set_cache_ptr(self._rd_cache, a)
-        cache = self._rd_cache.data
-        cache_hi = self._rd_cache.hi
-        cache_ptr = self._rd_cache.ptr
+        start, exact = self._get_cache_ptr(cache, a)
         # iterate over complete caches
-        while cache_hi <= b.date():
-            for data in cache[cache_ptr:]:
+        while cache.hi <= b.date():
+            for data in cache.data[start:]:
                 yield data
-            if cache_hi >= self._hi_limit:
+            if cache.hi >= self._hi_limit:
                 return
-            self._load(self._rd_cache, cache_hi)
-            cache = self._rd_cache.data
-            cache_hi = self._rd_cache.hi
-            cache_ptr = 0
+            self._load(cache, cache.hi)
+            start = 0
         # iterate over part of cache
-        for data in cache[cache_ptr:]:
+        for data in cache.data[start:]:
             if data['idx'] >= b:
                 return
             yield data
@@ -213,11 +213,11 @@ class CoreStore(object):
             return self._get_slice(i)
         if not isinstance(i, datetime):
             raise TypeError("list indices must be %s" % (datetime))
-        self._set_cache_ptr(self._rd_cache, i)
-        if (self._rd_cache.ptr >= len(self._rd_cache.data) or
-            self._rd_cache.data[self._rd_cache.ptr]['idx'] != i):
+        cache = self._rd_cache
+        ptr, exact = self._get_cache_ptr(cache, i)
+        if not exact:
             raise KeyError(i)
-        return self._rd_cache.data[self._rd_cache.ptr]
+        return cache.data[ptr]
 
     def __setitem__(self, i, x):
         """Store a value x with index i.
@@ -228,40 +228,33 @@ class CoreStore(object):
         if not isinstance(i, datetime):
             raise TypeError("index '%s' is not %s" % (i, datetime))
         x['idx'] = i
-        self._set_cache_ptr(self._wr_cache, i)
-        if len(self._wr_cache.data) == 0:
-            self._lo_limit = min(self._lo_limit, self._wr_cache.lo)
-            self._hi_limit = max(self._hi_limit, self._wr_cache.hi)
-            self._lo_limit_dt = datetime(
-                self._lo_limit.year, self._lo_limit.month, self._lo_limit.day)
-            self._hi_limit_dt = datetime(
-                self._hi_limit.year, self._hi_limit.month, self._hi_limit.day)
-        if (self._wr_cache.ptr < len(self._wr_cache.data) and
-            self._wr_cache.data[self._wr_cache.ptr]['idx'] == i):
-            self._wr_cache.data[self._wr_cache.ptr] = x
+        cache = self._wr_cache
+        ptr, exact = self._get_cache_ptr(cache, i)
+        if exact:
+            cache.data[ptr] = x
         else:
-            self._wr_cache.data.insert(self._wr_cache.ptr, x)
-        self._wr_cache.dirty = True
+            cache.data.insert(ptr, x)
+        cache.dirty = True
 
     def _del_slice(self, i):
         a, b = self._slice(i)
         if a > b:
             return
         # go to start of slice
-        self._set_cache_ptr(self._wr_cache, a)
+        cache = self._wr_cache
+        start, exact = self._get_cache_ptr(cache, a)
         # delete to end of cache
-        while self._wr_cache.hi <= b.date():
-            del self._wr_cache.data[self._wr_cache.ptr:]
-            self._wr_cache.dirty = True
-            if self._wr_cache.hi >= self._hi_limit:
+        while cache.hi <= b.date():
+            del cache.data[start:]
+            cache.dirty = True
+            if cache.hi >= self._hi_limit:
                 return
-            self._load(self._wr_cache, self._wr_cache.hi)
-            self._wr_cache.ptr = 0
+            self._load(cache, cache.hi)
+            start = 0
         # delete part of cache
-        ptr = self._wr_cache.ptr
-        self._wr_cache.set_ptr(b)
-        del self._wr_cache.data[ptr:self._wr_cache.ptr]
-        self._wr_cache.dirty = True
+        stop, exact = cache.get_ptr(b)
+        del cache.data[start:stop]
+        cache.dirty = True
 
     def __delitem__(self, i):
         """Delete the data item or items with index i.
@@ -272,12 +265,12 @@ class CoreStore(object):
             return self._del_slice(i)
         if not isinstance(i, datetime):
             raise TypeError("list indices must be %s" % (datetime))
-        self._set_cache_ptr(self._wr_cache, i)
-        if (self._wr_cache.ptr >= len(self._wr_cache.data) or
-            self._wr_cache.data[self._wr_cache.ptr]['idx'] != i):
+        cache = self._wr_cache
+        ptr, exact = self._get_cache_ptr(cache, i)
+        if not exact:
             raise KeyError(i)
-        del self._wr_cache.data[self._wr_cache.ptr]
-        self._wr_cache.dirty = True
+        del cache.data[ptr]
+        cache.dirty = True
 
     def before(self, idx):
         """Return datetime of newest existing data record whose
@@ -288,13 +281,14 @@ class CoreStore(object):
         if not isinstance(idx, datetime):
             raise TypeError("'%s' is not %s" % (idx, datetime))
         day = min(idx.date(), self._hi_limit - DAY)
+        cache = self._rd_cache
         while day >= self._lo_limit:
-            if day < self._rd_cache.lo or day >= self._rd_cache.hi:
-                self._load(self._rd_cache, day)
-            self._rd_cache.set_ptr(idx)
-            if self._rd_cache.ptr > 0:
-                return self._rd_cache.data[self._rd_cache.ptr - 1]['idx']
-            day = self._rd_cache.lo - DAY
+            if day < cache.lo or day >= cache.hi:
+                self._load(cache, day)
+            ptr, exact = cache.get_ptr(idx)
+            if ptr > 0:
+                return cache.data[ptr - 1]['idx']
+            day = cache.lo - DAY
         return None
 
     def after(self, idx):
@@ -306,32 +300,43 @@ class CoreStore(object):
         if not isinstance(idx, datetime):
             raise TypeError("'%s' is not %s" % (idx, datetime))
         day = max(idx.date(), self._lo_limit)
+        cache = self._rd_cache
         while day < self._hi_limit:
-            if day < self._rd_cache.lo or day >= self._rd_cache.hi:
-                self._load(self._rd_cache, day)
-            self._rd_cache.set_ptr(idx)
-            if self._rd_cache.ptr < len(self._rd_cache.data):
-                return self._rd_cache.data[self._rd_cache.ptr]['idx']
-            day = self._rd_cache.hi
+            if day < cache.lo or day >= cache.hi:
+                self._load(cache, day)
+            ptr, exact = cache.get_ptr(idx)
+            if ptr < len(cache.data):
+                return cache.data[ptr]['idx']
+            day = cache.hi
         return None
 
     def nearest(self, idx):
         """Return datetime of record whose datetime is nearest idx."""
         hi = self.after(idx)
+        if hi == idx:
+            return hi
         lo = self.before(idx)
         if hi is None:
             return lo
         if lo is None:
             return hi
-        if abs(hi - idx) < abs(lo - idx):
+        if (hi - idx) < (idx - lo):
             return hi
         return lo
 
-    def _set_cache_ptr(self, cache, i):
+    def _get_cache_ptr(self, cache, i):
         day = i.date()
         if day < cache.lo or day >= cache.hi:
             self._load(cache, day)
-        cache.set_ptr(i)
+            if cache.lo < self._lo_limit:
+                self._lo_limit = cache.lo
+                self._lo_limit_dt = datetime(
+                    cache.lo.year, cache.lo.month, cache.lo.day)
+            if cache.hi > self._hi_limit:
+                self._hi_limit = cache.hi
+                self._hi_limit_dt = datetime(
+                    cache.hi.year, cache.hi.month, cache.hi.day)
+        return cache.get_ptr(i)
 
     def _load(self, cache, target_date):
         self._flush(cache)
@@ -343,7 +348,6 @@ class CoreStore(object):
             cache.copy(self._rd_cache)
             return
         cache.data = []
-        cache.ptr = 0
         cache.path, cache.lo, cache.hi = new_path, new_lo, new_hi
         if not os.path.exists(cache.path):
             return
@@ -351,9 +355,14 @@ class CoreStore(object):
             kwds = {'mode': 'r', 'newline': ''}
         else:
             kwds = {'mode': 'rb'}
+        row_lengths = (len(self.key_list),
+                       len(self.key_list) - self.solar_items)
         with open(cache.path, **kwds) as csvfile:
             reader = csv.reader(csvfile, quoting=csv.QUOTE_NONE)
             for row in reader:
+                if len(row) not in row_lengths:
+                    logger.error('Invalid %s data at %s', self.dir_name, row[0])
+                    continue
                 result = {}
                 for key, value in zip(self.key_list, row):
                     if value == '':
@@ -443,6 +452,7 @@ class RawStore(CoreStore):
         'abs_pressure', 'wind_ave', 'wind_gust', 'wind_dir', 'rain',
         'status', 'illuminance', 'uv',
         ]
+    solar_items = 2
     conv = {
         'idx'          : WSDateTime.from_csv,
         'delay'        : int,
@@ -469,6 +479,7 @@ class CalibStore(CoreStore):
         'abs_pressure', 'rel_pressure', 'wind_ave', 'wind_gust', 'wind_dir',
         'rain', 'status', 'illuminance', 'uv',
         ]
+    solar_items = 2
     conv = {
         'idx'          : WSDateTime.from_csv,
         'delay'        : int,
@@ -496,6 +507,7 @@ class HourlyStore(CoreStore):
         'abs_pressure', 'rel_pressure', 'pressure_trend',
         'wind_ave', 'wind_gust', 'wind_dir', 'rain', 'illuminance', 'uv',
         ]
+    solar_items = 2
     conv = {
         'idx'               : WSDateTime.from_csv,
         'hum_in'            : int,
@@ -538,6 +550,7 @@ class DailyStore(CoreStore):
         'illuminance_ave', 'illuminance_max', 'illuminance_max_t',
         'uv_ave', 'uv_max', 'uv_max_t',
         ]
+    solar_items = 6
     conv = {
         'idx'                : WSDateTime.from_csv,
         'start'              : WSDateTime.from_csv,
@@ -633,6 +646,7 @@ class MonthlyStore(CoreStore):
         'uv_ave',
         'uv_max_lo', 'uv_max_lo_t', 'uv_max_hi', 'uv_max_hi_t', 'uv_max_ave',
         ]
+    solar_items = 12
     conv = {
         'idx'                  : WSDateTime.from_csv,
         'start'                : WSDateTime.from_csv,
